class FarmerHarvesterAPI:
    """A connected machine to the farmer

    schema of connections from example data
    {
        'bytes_read': 732920,
        'bytes_written': 736979,
        'creation_time': 1625781881.464225,
        'last_message_time': 1625856666.3932514,
        'local_port': 8447,
        'node_id': b'1n\x0f\xc4J\xb5q8\xc4\x98\x0b\xe7\\\xac\xd1\x82...',
        'peer_host': '127.0.0.1',
        'peer_port': 51844,
        'peer_server_port': 8448,
        'type': 2
    }
    """

    # pylint: disable=too-few-public-methods, too-many-instance-attributes

    bytes_read: int
    bytes_written: int
    creation_time: float
    last_message_time: float
    local_port: int
    node_id: bytes
    peer_host: str
    peer_port: int
    peer_server_port: int
    type: int
    n_plots: int

    def __init__(
        self,
        bytes_read: int,
        bytes_written: int,
        creation_time: float,
        last_message_time: float,
        local_port: int,
        node_id: bytes,
        peer_host: str,
        peer_port: int,
        peer_server_port: int,
        type: int,
        n_plots: int = 0,
    ):
        # pylint: disable=redefined-builtin, too-many-arguments
        self.bytes_read = bytes_read
        self.bytes_written = bytes_written
        self.creation_time = creation_time
        self.last_message_time = last_message_time
        self.local_port = local_port
        self.node_id = node_id
        self.peer_host = peer_host
        self.peer_port = peer_port
        self.peer_server_port = peer_server_port
        self.type = type
        self.n_plots = n_plots

    def copy(self) -> "FarmerHarvesterAPI":
        """Get a copy of the instance

        Returns
        -------
        farmer_harvester : FarmerHarvesterAPI
            copy of the instance
        """
        return FarmerHarvesterAPI(
            **self.__dict__,
        )



from datetime import datetime
from typing import Union


# pylint: disable=too-many-instance-attributes
class FarmerHarvesterLogfile:
    """Class with compact information about harvesters"""

    # General info
    harvester_id: str = ""
    ip_address: str = ""

    # Tracking of connection status
    is_connected = False

    # Signage point tracking
    time_of_end_of_last_sgn_point: Union[datetime, None]
    time_last_incoming_msg: Union[datetime, None]
    time_last_outgoing_msg: Union[datetime, None]

    # Metrics
    n_responses = 0
    n_overdue_responses = 0

    # Additional tracking
    last_update: datetime

    def __init__(
        self,
        harvester_id: str,
        ip_address: str,
        is_connected: bool = False,
        last_update: datetime = datetime.now(),
        time_of_end_of_last_sgn_point: Union[datetime, None] = None,
        time_last_incoming_msg: Union[datetime, None] = None,
        time_last_outgoing_msg: Union[datetime, None] = None,
        timed_out: bool = False,
        n_responses: int = 0,
        n_overdue_responses: int = 0,
    ):
        """Initializes the harvester info

        Parameters
        ----------
        harvester_id : str
            id of the harvester
        ip_address : str
            ip address of the harvester
        last_update: datetime
            timestamp when creation happened
        time_of_incoming_messages : Union[List[datetime], None]
            timestamps when messages came in from the harvester
        time_of_outgoing_messages : Union[List[datetime], None]
            timestamps when messages were sent to the harvester
        timed_out : bool
            if the harvester timed out
        time_of_timeout : Union[datetime, None]
            time of last timeout
        """
        # pylint: disable=too-many-arguments
        self.harvester_id = harvester_id
        self.ip_address = ip_address
        self.is_connected = is_connected
        self.last_update = last_update
        self.time_of_end_of_last_sgn_point = time_of_end_of_last_sgn_point
        self.time_last_incoming_msg = time_last_incoming_msg
        self.time_last_outgoing_msg = time_last_outgoing_msg
        self.timed_out = timed_out
        self.n_responses = n_responses
        self.n_overdue_responses = n_overdue_responses
        self.reset_times()

    def copy(self):
        """Get a copy of the HarvesterInfo

        Returns
        -------
        harvester_info : HarbesterInfo
            copy of this instance
        """
        return FarmerHarvesterLogfile(
            harvester_id=self.harvester_id,
            ip_address=self.ip_address,
            is_connected=self.is_connected,
            last_update=self.last_update,
            time_of_end_of_last_sgn_point=self.time_of_end_of_last_sgn_point,
            time_last_incoming_msg=self.time_last_incoming_msg,
            time_last_outgoing_msg=self.time_last_outgoing_msg,
            timed_out=self.timed_out,
            n_responses=self.n_responses,
            n_overdue_responses=self.n_overdue_responses,
        )

    def check_for_timeout(self, current_time: datetime) -> None:
        """This functions checks if the harvester timed out from the view of a farmer

        Parameters
        ----------
        current_time : datetime
            current time to base computation on

        Notes
        -----
            Modifies internal attributes so don't spam.
        """
        CHALLENGE_TIMEOUT = 25  # seconds
        HARVESTER_TIMOUT = 60  # seconds

        if not self.timed_out:
            if self.time_last_incoming_msg is not None and self.time_last_outgoing_msg is not None:
                delta_seconds = (current_time - self.time_last_incoming_msg).total_seconds()
                if delta_seconds > CHALLENGE_TIMEOUT:
                    self.n_overdue_responses += 1
                if delta_seconds > HARVESTER_TIMOUT:
                    self.timed_out = True

    def reset_times(self):
        """Resets time of incoming, outgoing msgs and signage points"""
        self.time_last_outgoing_msg = None
        self.time_last_incoming_msg = None
        self.time_of_end_of_last_sgn_point = None



from typing import Any, Dict, List

from chia.rpc.farmer_rpc_client import FarmerRpcClient
from chia.server.outbound_message import NodeType
from chia.util.config import load_config
from chia.util.default_root import DEFAULT_ROOT_PATH
from chia.util.ints import uint16

from ..ChiaWatchdog import ChiaWatchdog
from .FarmerHarvesterAPI import FarmerHarvesterAPI
from .shared_settings import API_EXCEPTIONS


async def _get_farmer_harvesters(
    farmer_client: FarmerRpcClient,
) -> List[FarmerHarvesterAPI]:

    harvesters = []

    # general connection information
    harvester_kwargs: Dict[str, Dict[str, Any]] = {
        kwargs["node_id"].hex(): kwargs
        for kwargs in await farmer_client.get_connections()
        if kwargs.get("type") == NodeType.HARVESTER.value
    }

    # harvester specific data such as plot count
    response = await farmer_client.get_harvesters()
    if response.get("success"):
        for kwargs in response.get("harvesters"):
            node_id = kwargs.get("connection").get("node_id")
            n_plots = len(kwargs.get("plots"))
            harvester_kwargs[node_id]["n_plots"] = n_plots

    for kwargs in harvester_kwargs.values():
        harvesters.append(FarmerHarvesterAPI(**kwargs))

    return harvesters


async def update_from_farmer(chia_dog: ChiaWatchdog):
    """Updates the chia dog with harvester data

    Parameters
    ----------
    chia_dog : ChiaWatchdog
        watchdog instance to be modified
    """
    # pylint: disable=duplicate-code

    try:
        config = load_config(DEFAULT_ROOT_PATH, "config.yaml", exit_on_error=False)
        self_hostname = config["self_hostname"]
        farmer_rpc_port = config["farmer"]["rpc_port"]
        farmer_client = await FarmerRpcClient.create(
            self_hostname,
            uint16(farmer_rpc_port),
            DEFAULT_ROOT_PATH,
            config,
        )

        chia_dog.farmer_service.connections = await _get_farmer_harvesters(
            farmer_client=farmer_client
        )
        chia_dog.farmer_service.is_running = True

    # pylint: disable=catching-non-exception
    except API_EXCEPTIONS:
        chia_dog.farmer_service.connections = []
        chia_dog.farmer_service.is_running = False
    finally:
        if "farmer_client" in locals():
            farmer_client.close()
            await farmer_client.await_closed()
        chia_dog.farmer_service.is_ready = True



from copy import deepcopy
from typing import Iterable, List


class HarvesterAPI:
    """This class holds chia information fetched through RPC
    from chia services on the same machine
    """

    # pylint: disable=too-few-public-methods

    # This helps us to check if an update happened
    is_ready: bool = False

    is_running: bool = False
    plots: List[dict]
    # Plot layout:
    # {
    #     'file_size': 108878195752,
    #     'filename': 'path/to/blabla.plot',
    #     'plot-seed': '0x00000000...',
    #     'plot_public_key': '0x00000000....',
    #     'pool_contract_puzzle_hash': None,
    #     'pool_public_key': '0x00000000....',
    #     'size': 32,
    #     'time_modified': 1621370658.446281
    # }
    failed_to_open_filenames: List[str]
    not_found_filenames: List[str]
    plot_directories: Iterable[str] = tuple()
    n_proofs: int = 0

    def __init__(self):
        self.plots = []
        self.failed_to_open_filenames = []
        self.not_found_filenames = []

    def copy(self) -> "HarvesterAPI":
        """Get a copy of the instance

        Returns
        -------
        harvester : HarvesterAPI
            copy of the instance
        """
        new_harvester = HarvesterAPI()
        new_harvester.is_running = self.is_running
        new_harvester.is_ready = self.is_ready
        new_harvester.plots = deepcopy(self.plots)
        new_harvester.failed_to_open_filenames = list(self.failed_to_open_filenames)
        new_harvester.not_found_filenames = list(self.not_found_filenames)
        new_harvester.plot_directories = list(self.plot_directories)
        new_harvester.n_proofs = self.n_proofs
        return new_harvester



class WalletAPI:
    """This class holds chia information fetched through RPC
    from chia services on the same machine
    """

    # pylint: disable=too-few-public-methods

    is_ready: bool = False
    n_wallets: int = 0
    is_synced: bool = False
    is_running: bool = False

    def copy(self) -> "WalletAPI":
        """Get a copy of the instance

        Returns
        -------
        wallet : WalletAPI
            copy of the instance
        """
        new_info = WalletAPI()
        new_info.n_wallets = self.n_wallets
        new_info.is_synced = self.is_synced
        new_info.is_ready = self.is_ready
        new_info.is_running = self.is_running
        return new_info



from typing import List
from .FarmerHarvesterAPI import FarmerHarvesterAPI


class FarmerAPI:
    """This class holds chia information fetched through RPC
    from chia services on the same machine
    """

    # pylint: disable=too-few-public-methods

    is_ready: bool = False
    is_running: bool = False
    connections: List[FarmerHarvesterAPI]

    def __init__(self):
        self.connections = []

    def copy(self) -> "FarmerAPI":
        """Get a copy of the instance

        Returns
        -------
        harvester : FarmerAPI
            copy of the instance
        """
        new_farmer = FarmerAPI()
        new_farmer.is_ready = self.is_ready
        new_farmer.is_running = self.is_running
        new_farmer.connections = [connection.copy() for connection in self.connections]
        return new_farmer



import asyncio
from typing import Dict, List

from .api.FarmerAPI import FarmerAPI
from .api.HarvesterAPI import HarvesterAPI
from .api.WalletAPI import WalletAPI
from .logfile.FarmerHarvesterLogfile import FarmerHarvesterLogfile
from .madmax.MadMaxPlotInProgress import MadMaxPlotInProgress


class ChiaWatchdog:
    """Class for watching chia"""

    # pylint: disable=too-many-instance-attributes

    __logfile_chia_ready: bool = False
    __logfile_madmax_ready: bool = False

    # members related to logfile checking
    harvester_infos: Dict[str, FarmerHarvesterLogfile]
    farmed_blocks: List[str]

    # members related to madmax plotter
    plots_in_progress: List[MadMaxPlotInProgress]

    # members for contacting chia directly
    farmer_service: FarmerAPI
    wallet_service: WalletAPI
    harvester_service: HarvesterAPI

    def __init__(self, logfile_filepath: str, madmax_logfile: str):
        """initialize a chia watchdog

        Parameters
        ----------
        logfile_filepath : str
            path to the logfile to watch
        madmax_logfile : str
            path to the madmax logfile to watch
        """
        self.logfile_filepath = logfile_filepath
        self.madmax_logfile = madmax_logfile
        self.harvester_infos = {}
        self.farmer_service = FarmerAPI()
        self.wallet_service = WalletAPI()
        self.harvester_service = HarvesterAPI()
        self.farmed_blocks = []
        self.plots_in_progress = []

    def copy(self) -> "ChiaWatchdog":
        """Creates a copy of this instance

        Returns
        -------
        new_instance : ChiaWatchdog
            copy of the instance
        """
        new_instance = ChiaWatchdog(self.logfile_filepath, self.madmax_logfile)
        new_instance.harvester_infos = {
            id: harvester_info.copy() for id, harvester_info in self.harvester_infos.items()
        }
        new_instance.plots_in_progress = [plot.copy() for plot in self.plots_in_progress]
        # not sure about this one but ok
        # pylint: disable=protected-access
        # pylint: disable=unused-private-member
        new_instance.__logfile_chia_ready = self.__logfile_chia_ready
        new_instance.__logfile_madmax_ready = self.__logfile_madmax_ready
        new_instance.farmer_service = self.farmer_service.copy()
        new_instance.wallet_service = self.wallet_service.copy()
        new_instance.harvester_service = self.harvester_service.copy()
        new_instance.farmed_blocks = self.farmed_blocks.copy()

        return new_instance

    async def ready(self):
        """Wait for the readiness of the watchdog"""
        while not (
            self.__logfile_chia_ready
            and self.__logfile_madmax_ready
            and self.harvester_service.is_ready
            and self.farmer_service.is_ready
            and self.wallet_service.is_ready
        ):
            await asyncio.sleep(0.25)

    def set_chia_logfile_is_ready(self):
        """When the chia logfile scanner has done its init, this gets called"""
        self.__logfile_chia_ready = True

    def set_madmax_logfile_is_ready(self):
        """When the madmax logfile scanner has done its init, this gets called"""
        self.__logfile_madmax_ready = True

    def get_or_create_harvester_info(
        self,
        harvester_id: str,
        ip_address: str,
    ) -> FarmerHarvesterLogfile:
        """Get or create a harvester info

        Parameters
        ----------
        harvester_id : str
            id of the harvester
        ip_address : str
            ip address

        Returns
        -------
        harvester_info : FarmerHarvesterLogfile
            existing or newly created harvester info
        """

        harvester_info = self.harvester_infos.get(harvester_id)

        #  does not exist yet
        if harvester_info is None:
            harvester_info = FarmerHarvesterLogfile(
                harvester_id=harvester_id,
                ip_address=ip_address,
            )
            self.harvester_infos[harvester_id] = harvester_info
        else:
            # make sure ip is updated
            harvester_info.ip_address = ip_address

        return harvester_info
